from mercurial.node import hex
from mercurial import bookmarks
from mercurial import registrar

cmdtable = {}
command = registrar.command(cmdtable)


def _collectdraftcommits(node):
    queue = [node]
    drafts = set()
    enqueued = set()
    while queue:
        node = queue.pop(0)
        id = hex(node.node())
        if node.phase() == 0 or node in drafts:
            continue

        drafts.add(id)
        new = list()
        for p in node.parents():
            if not p in enqueued and p.phase() != 0:
                new.append(p)
                enqueued.add(p)
        queue = sorted(queue + new, key=lambda node: -node.rev())
    return drafts


def normalizename(name):
    return name.replace('\\', '\\\\').replace('\n', '\\n').replace('\t', '\\t')


@command('sgbranches')
def sgbranches(ui, repo):
    ctx = repo[None]
    parents = [p for p in ctx.parents() if p.rev() >= 0]
    ui.write('heads\t', str(len(parents)), '\n')
    for p in parents:
        ui.write(hex(p.node()), '\n')
    ui.write('currentbranch\t', normalizename(ctx.branch()), '\n')

    determinedraftcount = len(ui.configitems("paths")) > 0
    activebranches = set([repo[n].branch() for n in repo.heads()])
    branchmap = dict(repo.branchmap())
    for branch in activebranches:
        if not branch in branchmap:
            branchmap[branch] = list()

    ui.write('branches\t', str(len(branchmap)), '\n')
    for ref, heads in branchmap.iteritems():
        alldraftcommits = set()
        tuples = list()
        for h in reversed(heads):
            head = repo[h]
            draftcommits = _collectdraftcommits(head) if determinedraftcount else list()
            tuples.append((head, len(draftcommits)))
            alldraftcommits = alldraftcommits.union(draftcommits)

        ui.write(normalizename(ref), '\t', str(len(heads)), '\t', str(len(alldraftcommits)), '\n')
        for tuple in tuples:
            head = tuple[0]
            draftcommits = tuple[1]
            isopen = not head.closesbranch()
            isactive = ref in activebranches and isopen
            ui.write(str(head.rev()), '\t', hex(head.node()), '\t', 'open' if isopen else 'closed', '\t', 'active' if isactive else 'inactive', '\t',
                str(draftcommits), '\n')

    taginfos = list()
    for ref, n in reversed(repo.tagslist()):
        rev = repo.changelog.rev(n)
        if rev >= 0:
            taginfos.append((ref, repo.changelog.rev(n), hex(n), repo.tagtype(ref)))

    ui.write('tags\t', str(len(taginfos)), '\n')
    for ref, rev, sha, type in reversed(taginfos):
        ui.write(normalizename(ref), '\t', str(rev), '\t', sha, '\t', str(type), '\n')

    marks = bookmarks.bmstore(repo)
    curMark = marks.active

    ui.write('activebookmark\t', curMark if curMark is not None else '', '\n')
    ui.write('bookmarks\t', str(len(marks)), '\n')
    for bmark, n in sorted(marks.iteritems()):
        h = repo[n]
        draftcommits = len(_collectdraftcommits(h)) if determinedraftcount else 0
        ui.write(normalizename(bmark), '\t', str(repo.changelog.rev(n)), '\t', hex(h.node()), '\t', str(draftcommits), '\n')
